/* Copyright (c) Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gtk/gtk.h>
#include <X11/Xatom.h>
#include <gdk/gdkx.h>
#include <vte/vte.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <mntent.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <libintl.h>
#include "splash.h"
#include "config.h"
#include "set-keymap.h"
#include <gdk-pixbuf/gdk-pixdata.h>
#include <libintl.h>
#include <popt.h> 
#include <sys/mount.h>

#ifndef MNT_DETACH
#define MNT_DETACH 0x00000002
#endif

#define TMPPATH "/etc/rhgb/temp"
#define SOCKET_NAME TMPPATH"/rhgb-socket"
#define DISPLAY_NO ":9"
#define LOCALHOST "127.0.0.1"
#define DISPLAY_STRING LOCALHOST DISPLAY_NO
#define DISPLAY_FNAME TMPPATH"/display"
#define PIPE_NAME TMPPATH"/rhgb-console"
#define X_CONFIG "/etc/rhgb/xorg.conf"
#define XLOGFILE TMPPATH"/xorg.log"
#define DEBUGFILE TMPPATH"/rhgb.debug"
#define X_NO_DRI "-extension XFree86-DRI"
#define VT_NAME "/dev/tty8"
#define X_COMMAND1 "/usr/bin/Xorg -s off -dpms -v -nolock -fp /usr/share/X11/fonts/misc -logfile " XLOGFILE " " X_NO_DRI " " DISPLAY_NO " vt8"
#define X_COMMAND2 "/usr/bin/Xorg -config " X_CONFIG " -s off -dpms -v -nolock -fp /usr/share/X11/fonts/misc -logfile " XLOGFILE " " X_NO_DRI " " DISPLAY_NO " vt8"
#define NETWORK_COMMAND "/sbin/ifconfig lo " LOCALHOST " up"
#define DISPLAY_ENV "DISPLAY=" DISPLAY_STRING
#define FIX_INIT_COMMAND "/sbin/change_console -r"
#define SYSRESOURCES "/etc/X11/Xresources"
#define XRDB "/usr/bin/xrdb"
#define XRDB_COMMAND XRDB " -display " DISPLAY_STRING " -merge " SYSRESOURCES

/*
 * Maximum time to wait until the sysinit finished and we were asked to quit
 */
#define QUIT_TIMEOUT 30

/* Keep in sync w/ rhgb-client.c */
#define N_STEPS 21

/* For localizations with gettext */
#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) (x)


typedef struct _ServiceMapping ServiceMapping;
struct _ServiceMapping
{
  const char *service;
  const char *readable_name;
};

ServiceMapping service_list[] =
  {
    {"RCclock", N_("Setting up system clock")},
    {"RChostname", N_("Setting the hostname")},
    {"RCmountfs", N_("Setting up local disks")},
    {"random", N_("Starting random number generator")},
    {"gpm", N_("Starting console mouse services")},
    {"kudzu", N_("Probing for new hardware")},
    {"pcmcia", N_("Starting PCMCIA card services")},
    {"iptables", N_("Starting firewall")},
    {"ipchains", N_("Starting firewall")},
    {"apmd", N_("Starting power management services")},
    {"cups", N_("Starting printing system")},
    {"isdn", N_("Starting isdn")},
    {"network", N_("Starting networking")},
    {"sshd", N_("Starting sshd")},
    {"ntpd", N_("Starting time server")},
    {"rhnsd", N_("Starting Red Hat Network support")},
    {"sendmail", N_("Starting E-Mail transport")},
    {"spamassassin", N_("Starting E-Mail spam filter system")},
    {"firstboot", N_("Starting first-time boot configuration")},
    {"loginscreen", N_("Starting login services")}
  };

static gint n_intervals;
static gint current_interval = -1;
static gboolean force_arg = 0;
static gboolean debug = 0;
static gboolean force_debug = 0;
static gboolean debug_interface = 0;
static char *use_existing_server = NULL;

static gint child_pid = 0;
static GList *services = NULL;
static Splash *splash = NULL;
/* Timeouts */
static guint motion_timeout = 0;
static guint motion_notify_event_id = 0;

static gint socket_fd = -1;
static gint pipe_fd = -1;
static sig_atomic_t display_is_ready = FALSE;
static gint display_ready_fds[2] = { -1, -1 };

static gint exit_code = 0;

static const char *system_service_names[] =
{
  "RCconsole",
  "RCproc",
  "RCkernelparam",
  "RCclock",
  "RChostname",
  "RCusb",
  "RCmountfs",
  "RCswap",
  "RCraid"
};

static gboolean unmount_ramfs (void);

static void
DEBUG_MESSAGE (const char *message, ...)
{
  gchar *full_message;
  int dbg;
  va_list args;

  if ((debug == 0) && (force_debug == 0))
    return;

  va_start (args, message);
  full_message = g_strdup_vprintf (message, args);
  va_end (args);
  if (full_message == NULL)
      return;

  g_print (full_message);
  /* synchronous write of the debug logs */
  dbg = open(DEBUGFILE, O_WRONLY | O_CREAT | O_APPEND, 0644);
  if (dbg >= 0) {
      write(dbg, full_message, strlen(full_message));
      close(dbg);
  }
  g_free (full_message);
}

/*
 * The shutdown function. Kill the specific X process, and umount
 * the special filesystem
 */
static void
rhgb_shutdown(void)
{
    if (child_pid > 0) {
        kill (child_pid, SIGTERM);
	child_pid = 0;
    }
    if (socket_fd >= 0) {
	close(socket_fd);
	socket_fd = -1;
    }
    if (pipe_fd >= 0) {
	close(pipe_fd);
	pipe_fd = -1;
    }
    unlink(SOCKET_NAME);
    unlink(XLOGFILE);
    unlink(DISPLAY_FNAME);
    unlink(PIPE_NAME);
    unlink(DEBUGFILE);
    unmount_ramfs();
    system(FIX_INIT_COMMAND);
}

/*
 * If we segfault, try to cleanup
 */
static void
got_signal(int no ATTRIBUTE_UNUSED) {
    if (child_pid > 0) {
        kill (child_pid, SIGTERM);
    }
    if (socket_fd >= 0) {
	close(socket_fd);
    }
    if (pipe_fd >= 0) {
	close(pipe_fd);
    }
    unlink(SOCKET_NAME);
    unlink(XLOGFILE);
    unlink(DISPLAY_FNAME);
    unlink(PIPE_NAME);
    unlink(DEBUGFILE);
    unmount_ramfs();
    system(FIX_INIT_COMMAND);
    _exit(1);
}

/*
 * Timeout handling. If we don't get pings or progress informations
 * for a while we switch back automatically to text mode console
 * connections_timeout is reset to 0 each time we get some client command
 * If the boot process is stuck into some long command (fsck, network
 * or DNS start error) then it will automatically switch to the console mode
 * to show what is happening.
 */
static guint pipe_timeout = 0;
/*
 * If this is set to a negative number, we are supposed to
 * not timeout (e.g. for firstboot)
 */
static int connection_timeout = 0;
static int switched_vt1 = 0;
static int switched_details = 0;
/*
 * Set if the user manually switched to details view, so we don't automatically
 * switch back in that case.
 */
static gboolean manual_switch = 0;

static void 
connections_reset (gboolean from_pipe) 
{
    pipe_timeout = 0;
    if (switched_vt1) {
        switch_to_vt (8);
	switched_vt1 = 0;
    }
    if (!from_pipe && connection_timeout >= 0)
	connection_timeout = 0;
    if ((!from_pipe) && (switched_details) && (!manual_switch)) {
        /* switch back to no details */
	splash_hide_details(splash);
	switched_details = 0;
    }
}

static gboolean
connections_check (void) 
{
    if(connection_timeout >= 0) connection_timeout++;
    pipe_timeout++;
    /*
     * Those checks are intended to detect stops in progresses of the
     * boot and switch back/forth to/from the detailed or text views.
     */
    if ((connection_timeout >= 10) && (switched_vt1 == 0) &&
        (switched_details == 0)) {
        /*
	 * Something is blocking could be filesystem check or kudzu
	 * switch to detailed mode
	 */
        splash_show_details(splash);
	switched_details = 1;
    }

    return (TRUE);
}

static gboolean
init_timeout (void) {
    static int initialized = 0;

    if (initialized)
        return(FALSE);
    g_timeout_add(1000, (GSourceFunc) connections_check, NULL);
    initialized = 1;
    return(TRUE);
}

/* Check for x
 */
static gboolean
check_for_x (void)
{
  return (access ("/usr/bin/Xorg", R_OK | X_OK));
}

static gboolean
check_for_xrdb (void)
{
  if (!access(SYSRESOURCES, R_OK))
      return(FALSE);
  return (access (XRDB, R_OK | X_OK));
}

/* Get the language
 */
static gchar *
check_for_language (void)
{
#if (GLIB_MAJOR_VERSION > 2) || ((GLIB_MAJOR_VERSION == 2) && \
    (GLIB_MINOR_VERSION >= 4))
  gchar *contents, **words;
  gchar *ret = NULL;
  gsize len;
  int x;

  g_file_get_contents("/etc/sysconfig/keyboard",&contents,&len, NULL);
  if (!contents)
    return(NULL);

  words = g_strsplit_set(contents,"\n\r \t='\"_-",0);
  for (x = 0; words[x] ; x++)
    {
       if (!strcmp(words[x],"KEYTABLE"))
         {
	   x++;
	   while ((words[x] != NULL) && ((words[x][0] == 0))) x++;
	   ret = g_strdup(words[x]);
	   break;
	 }
    }
  g_free(contents);
  g_strfreev(words);
  return(ret);
#else
  return(NULL);
#endif
}

/* Check that we should start (i.e., that we are booting runlevel 5)
 */
static gboolean
check_for_run (void)
{
  gchar *contents, **args;
  gsize len;
  int x,rc = -1;
  gboolean rhgb_found = FALSE;
	

  /* Check that an alternate runlevel (other than 5) wasn't specified
   * on the command line */
  g_file_get_contents("/proc/cmdline",&contents,&len, NULL);
  if (!contents)
    return 1;
  contents = g_strstrip(contents);
  args = g_strsplit(contents," ",0);
  for (x = 0; args[x] ; x++)
    {
      if (g_ascii_isdigit(args[x][0]))
	{
	  if (atoi(args[x]) != 5)
	    rc = 1;
	  else
	    rc = 0;
	  break;
	}
      if (!strcmp(args[x],"-s") ||
	  !strcasecmp(args[x],"S") ||
	  !strcmp(args[x],"single"))
	{
	  rc = 1;
	  break;
	}
      if (!strcmp(args[x],"rhgb"))
	rhgb_found = TRUE;
    }
  g_free(contents);
  g_strfreev(args);
  if (rc >= 0) return rc;
  if (!rhgb_found) return 1;

  /* Check that the default runlevel is 5 */	
  g_file_get_contents("/etc/inittab",&contents,&len, NULL);
  if (!contents)
    return 1;
  args = g_strsplit(contents,"\n",0);
  for (x = 0; args[x] ; x++) {
    args[x] = g_strchug(args[x]);
    if (!strncmp(args[x],"id:",3))
      {
	if (atoi(args[x]+3) != 5)
	  rc = 1;
	else
	  rc = 0;
      }
  }
  g_free(contents);
  g_strfreev(args);
  return rc;
}

/*
 * Set up RAMfs for temporary files
 */
static gboolean
setup_ramfs(void)
{
    FILE *mtab;
    struct mntent *mountpoint;
    gboolean flag = FALSE;

    mtab = setmntent("/etc/mtab", "r");
    if (mtab == NULL)
        return 1;
    while ((mountpoint = getmntent(mtab)) != NULL) {
        if ((!strcmp(mountpoint->mnt_dir, TMPPATH)) &&
            (!strcmp(mountpoint->mnt_type, "ramfs"))) {
            flag = TRUE;
            break;
        }
    }
    endmntent(mtab);

    if (flag)
        return 0;

    if (mount("none", TMPPATH, "ramfs", 0, "maxsize=512"))
        return 1;

    chmod(TMPPATH, S_IRUSR | S_IWUSR | S_IXUSR);
    /* chdir(TMPPATH); */
    return 0;
}

/*
 * Unmount RAMfs
 */
static gboolean unmount_ramfs (void)
{
  return umount2(TMPPATH,MNT_DETACH);
}


/* Utility code
 */
static void
setup_failed (void)
{
  /* Do something */
  if (child_pid > 0)
    kill (child_pid, SIGTERM);
  close(socket_fd);
  close(pipe_fd);
  unlink(SOCKET_NAME);
  unlink(XLOGFILE);
  unlink(DISPLAY_FNAME);
  unlink(PIPE_NAME);
  unmount_ramfs();
  switch_to_vt (1);
  system(FIX_INIT_COMMAND);
  _exit (exit_code);
}

static gboolean
virtual_console_init (void)
{
  int fd;
  gboolean console_initialized;

  console_initialized = FALSE;
  fd = open (VT_NAME, O_RDWR | O_NOCTTY, 0);

  /* initialize console for utf-8 mode if running in utf-8 locale
   */
  if ((fd >= 0) && g_get_charset (NULL) /* TRUE if UTF-8 */)
    {
      static const char utf8_enable_command[] = "\033%G";
      ssize_t bytes_written;
      char *set_font_command;

      bytes_written = write (fd, utf8_enable_command,
                             sizeof (utf8_enable_command));
      close (fd);

      if (bytes_written != sizeof (utf8_enable_command))
        console_initialized = FALSE;
      else
        {
          const char *font;
          
          font = g_getenv ("SYSFONT");

          if (font == NULL)
            font = "latarcyrheb-sun16";

          set_font_command = g_strdup_printf ("setfont -C %s %s",
                                              VT_NAME, font);

          if (!g_spawn_command_line_sync (set_font_command, NULL, NULL, NULL, NULL))
            console_initialized = FALSE;
          else
            console_initialized = TRUE;

            g_free (set_font_command);
	}
    }
  else
    console_initialized = FALSE;

  return console_initialized;
}

static void
child_setup (void)
{
    signal (SIGUSR1, SIG_IGN);
}

static void
on_display_ready (int signal)
{
    display_is_ready = TRUE;
    close (display_ready_fds[1]);
    display_ready_fds[1] = -1;
}

/* Display
 */

static void
display_init (void)
{
  gchar **argv;
  GIOChannel *channel;
  int fd;

  if (use_existing_server != NULL)
    {
      gchar *env;

      env = g_strdup_printf ("DISPLAY=%s", use_existing_server);
      putenv (env);
      return;
    }

  system (NETWORK_COMMAND);
  unsetenv ("XAUTHORITY");
  putenv ((char *) DISPLAY_ENV);
  
  fd = open(X_CONFIG, O_RDONLY);
  if (fd >= 0) {
      close(fd);
      g_shell_parse_argv (X_COMMAND2,
			  NULL, &argv,
			  NULL);
  } else {
      g_shell_parse_argv (X_COMMAND1,
			  NULL, &argv,
			  NULL);
  }

  /* ignore error.  It shouldn't really be fatal I think
   * and rhgb doesn't do any sort of logging, unfortunately.
   */
  virtual_console_init ();

  if (pipe (display_ready_fds) < 0) {
	  display_ready_fds[0] = -1;
	  display_ready_fds[1] = -1;
	  exit_code = 2;
	  return;
  }

  fcntl (display_ready_fds[0], F_SETFD, FD_CLOEXEC);
  fcntl (display_ready_fds[1], F_SETFD, FD_CLOEXEC); 

  channel = g_io_channel_unix_new (display_ready_fds[0]);
  g_io_add_watch (channel, 
		  G_IO_HUP | G_IO_ERR,
		  (GIOFunc) gtk_main_quit, NULL);
  g_io_channel_unref (channel);
  signal (SIGUSR1, on_display_ready);
  g_spawn_async ("/", argv, NULL,
		 G_SPAWN_DO_NOT_REAP_CHILD, 
		 (GSpawnChildSetupFunc) child_setup, NULL, &child_pid, NULL);
  g_child_watch_add (child_pid, (GChildWatchFunc) gtk_main_quit, NULL);

  gtk_main ();

  if (display_ready_fds[0] >= 0)
	  close (display_ready_fds[0]);
  if (display_ready_fds[1] >= 0)
	  close (display_ready_fds[1]);

  g_strfreev (argv);

  if (!display_is_ready)
	  return;

  fd = open(DISPLAY_FNAME, O_WRONLY | O_CREAT , 0644);
  if (fd >= 0) {
      write(fd, DISPLAY_STRING, strlen(DISPLAY_STRING));
      close(fd);
  }
}

static gboolean
merge_resources (void)
{
  system (XRDB_COMMAND);
  return 0;
}

static gboolean
hide_main_window_timeout_cb (gpointer data ATTRIBUTE_UNUSED)
{
  splash_hide_screen (splash);

  return FALSE;
}

static void
update_progress_bar (const char *step)
{
  GList *list;
  gint count = 0;
  gchar *text;
  guint i;

  for (list = services; list; list = list->next)
    {
      if (! strcmp (step, list->data))
	{
	  if (count > current_interval)
	    {
	      text = NULL;
	      for (i = 0; i < G_N_ELEMENTS (service_list); i++)
		{
		  if (!strcmp (step, service_list[i].service))
		    {
		      text = _(service_list[i].readable_name);
		    }
		  /* We know firstboot will be the last step, so we give it 6
		   * seconds to popup and then hide the window.  We also use a
		   * normal cursor.
		   * We also try to set the keyboard language since the value
		   * specified in the X config file is not taken into account
		   * on a read-only filesystem, the keymap is generated into
		   * file and then loaded by X, which fails when we start 
		   * rhgb, use setxkbmap to set it up.
		   */
		  if (!strcmp (step, "firstboot"))
		    {
		      gchar *lang;
		      GdkCursor *cursor;

                      connection_timeout = -1;
                      lang = check_for_language();
		      if (lang != NULL) {
		          gchar *argv[5] = {
			      (gchar *) "/usr/bin/setxkbmap",
			      (gchar *) "-display",
			      (gchar *) DISPLAY_NO,
			      NULL,
			      NULL
			  };
			  argv[3] = lang;
			  g_spawn_async(NULL, &argv[0], NULL, 
			                0, NULL, NULL, NULL, NULL);
		          g_free(lang);
		      }

		      /*
		       * then try more accurate but more likely to fail method
		       */
		      set_xkb_keymap();

		      g_timeout_add (6001, hide_main_window_timeout_cb, NULL);
		      if (motion_notify_event_id)
			{
			  motion_notify_event_id = 0;
			}
		      if (motion_timeout)
			{
			  g_source_remove (motion_timeout);
			  motion_timeout = 0;
			}

		      cursor = gdk_cursor_new (GDK_LEFT_PTR);
		      gdk_window_set_cursor (gdk_get_default_root_window (), cursor);
		      gdk_cursor_unref (cursor);
		    }
		}

	      current_interval = count;

	      if (n_intervals == 0)
		splash_pulse (splash);
	      else
		splash_set_progress (splash, (count+1)/(float)n_intervals);
	      /* We set the text after the progress so that the setting the
	       * progressbar doesn't clear it
	       */
	      if (text)
		{
		  splash_set_label (splash, text);
		}
	    }
	  break;
	}
      count++;
    }
}

/* Communication with the client
 */
static void
handle_command (char *command, gint return_fd)
{
  DEBUG_MESSAGE ("got a command:%s:\n", command);

  if (! strcmp (command, "quit"))
    {
      gtk_main_quit ();
      write (return_fd, "done", strlen("done") + 1);
      update_progress_bar("loginscreen");
    }
  else if (! strcmp (command, "ping"))
    {
      write (return_fd, "pong", strlen ("pong") + 1);
    }
  else if (! strncmp (command, "update=", strlen ("update=")))
    {
      update_progress_bar (command + strlen ("update=")); 
    }
  else if (!strncmp (command, "details", strlen ("details")))
    {
      manual_switch = TRUE;
      splash_show_details (splash);
    }
  else if (!strncmp (command, "nodetails", strlen ("nodetails")))
    {
      manual_switch = FALSE;
      splash_hide_details (splash);
    }
  else if (!strncmp (command, "sysinit", strlen ("sysinit")))
    {
      /* set the keyboard map */
      set_xkb_keymap ();
    }
  close (return_fd);
}

static gboolean
client_message_func (GIOChannel   *source,
		     GIOCondition  condition ATTRIBUTE_UNUSED,
		     gint          return_fd)
{
  gchar buffer[200];
  gsize bytes_read = 0;
  GIOStatus status;

  /* FIXME: We expect the command to be less than 200 bytes, come in one read,
   * and be NULL terminated.  We ignore everything that's not like that.
   */
  status = g_io_channel_read_chars (source,
				    buffer,
				    199,
				    &bytes_read,
				    NULL);

  switch (status)
    {
    case G_IO_STATUS_NORMAL:
      connections_reset (FALSE);
      if (bytes_read >= 200)
	{
	  DEBUG_MESSAGE ("More than 200 bytes: %d\n", bytes_read);
	  goto done;
	}
      /* Make sure it's NULL terminated */
      buffer[bytes_read] = '\000';
      DEBUG_MESSAGE ("got the buffer:%d:%s:\n", bytes_read, buffer);
      handle_command (buffer, return_fd);
      goto done;
    case G_IO_STATUS_AGAIN:
      DEBUG_MESSAGE ("G_IO_STATUS_AGAIN\n");
      return TRUE;
    case G_IO_STATUS_EOF:
      DEBUG_MESSAGE ("G_IO_STATUS_EOF\n");
      break;
    case G_IO_STATUS_ERROR:
      DEBUG_MESSAGE ("G_IO_STATUS_ERROR\n");
      break;
    default:
      g_assert_not_reached ();
    }

 done:
  close (g_io_channel_unix_get_fd (source));
  return FALSE;
}


static gboolean
socket_func (GIOChannel   *source,
	      GIOCondition  condition,
	      gpointer      data ATTRIBUTE_UNUSED)
{
  if (condition == G_IO_IN)
    {
      GIOChannel *new_channel;
      gint new_fd;

      new_fd = accept (g_io_channel_unix_get_fd (source), NULL, 0);

      new_channel = g_io_channel_unix_new (new_fd);
      g_io_channel_set_encoding (new_channel, NULL, NULL);
      g_io_channel_set_flags (new_channel,
			      g_io_channel_get_flags (new_channel) | G_IO_FLAG_NONBLOCK, NULL);
      g_io_add_watch (new_channel, G_IO_IN, (GIOFunc) client_message_func, GINT_TO_POINTER (new_fd));
      g_io_channel_unref (new_channel);
    }
  return TRUE;
}

/*
 * create the socket to communicate with application rhgb_client
 * the socket is under SOCKET_NAME i.e. /etc/rhgb/temp/rhgb-socket
 */
static gboolean
socket_init (void)
{
  int curstate = 1;
  struct sockaddr_un *saddr; 
  GIOChannel *sock_channel;

  socket_fd = socket (PF_UNIX, SOCK_STREAM, 0);

  if (socket_fd < 0)
    return -1;

  if (setsockopt (socket_fd, SOL_SOCKET, SO_REUSEADDR, &curstate, sizeof(curstate)) < 0)
    return -1;

  /*
   * Make sure socket closed when exec commands happen
   */
  fcntl(socket_fd, F_SETFD, FD_CLOEXEC);

  saddr = g_new0 (struct sockaddr_un, 1);
  saddr->sun_family = AF_UNIX;
  memcpy (saddr->sun_path, SOCKET_NAME, sizeof (SOCKET_NAME));

  if (bind (socket_fd, (struct sockaddr *) saddr, sizeof (struct sockaddr_un)))
    return -1;

  if (listen (socket_fd, 5))
    return -1;

  chmod (SOCKET_NAME, S_IRUSR | S_IWUSR | S_IXUSR);

  sock_channel = g_io_channel_unix_new (socket_fd);
  g_io_add_watch (sock_channel, G_IO_IN, socket_func, NULL);
  g_io_channel_unref (sock_channel);

  return 0;
}

/*
 * create the fifo to echo data before init process console is reset
 * needed for rc.sysinit
 * the fifo is under PIPE_NAME i.e. /etc/rhgb/temp/rhgb-console
 */
static gboolean
pipe_func(GIOChannel * source,
          GIOCondition condition, gpointer data ATTRIBUTE_UNUSED)
{
    if (condition == G_IO_IN) {
        gchar buffer[4097];
        gsize bytes_read = 0;
        GIOStatus status;
        status = g_io_channel_read_chars(source,
                                         buffer, 4096, &bytes_read, NULL);

        switch (status) {
            case G_IO_STATUS_NORMAL:
                connections_reset (TRUE);
                if (bytes_read > 4096) {
                    DEBUG_MESSAGE("More than 4096 bytes: %d\n", bytes_read);
                    goto done;
                }
                /* Make sure it's NULL terminated */
                buffer[bytes_read] = '\000';
                DEBUG_MESSAGE("got the pipe:%d:%s:\n", bytes_read,
                              buffer);
                splash_echo(splash, buffer, (glong) bytes_read);
                return TRUE;
            case G_IO_STATUS_AGAIN:
                DEBUG_MESSAGE("G_IO_STATUS_AGAIN\n");
                return TRUE;
            case G_IO_STATUS_EOF:
                DEBUG_MESSAGE("G_IO_STATUS_EOF\n");
                break;
            case G_IO_STATUS_ERROR:
                DEBUG_MESSAGE("G_IO_STATUS_ERROR\n");
                break;
            default:
                g_assert_not_reached();
        }

      done:
        close(g_io_channel_unix_get_fd(source));
	pipe_fd = -1;
        return FALSE;
    }
    return TRUE;
}

static gboolean
pipe_init (void)
{
    int ret;
    GIOChannel *pipe_channel;

    unlink(PIPE_NAME);

    ret = mkfifo(PIPE_NAME, S_IRUSR | S_IWUSR);
    if (ret != 0)
        return -1;

    pipe_fd = open(PIPE_NAME, O_NONBLOCK | O_RDWR);
    if (pipe_fd < 0)
        return -1;

    /*
     * Make sure the pipe is when exec commands happen
     */
    fcntl(pipe_fd, F_SETFD, FD_CLOEXEC);

    pipe_channel = g_io_channel_unix_new(pipe_fd);
    g_io_channel_set_encoding(pipe_channel, NULL, NULL);
    g_io_add_watch(pipe_channel, G_IO_IN, pipe_func, NULL);
    g_io_channel_unref(pipe_channel);
    return 0;
}


static const char *invalid_suffix[] = 
{
  ".rpmsave",
  ".rpmorig",
  ".rpmnew",
  "~",
  ".orig"
};

static gboolean
check_for_invalid_service (const char *name)
{
  guint i;
  gint name_len = strlen (name);

  g_assert (name);

  if (name[0] != 'S')
    return TRUE;

  if (name_len < 3)
    return TRUE;
  
  for (i = 0; i < G_N_ELEMENTS (invalid_suffix); i++)
    {
      const gchar *suffix;
      gint suffix_len;

      suffix = invalid_suffix[i];
      suffix_len = strlen (suffix);

      if (name_len <= suffix_len)
	continue;

      if (! strcmp (name + name_len - suffix_len, suffix))
	return TRUE;
    }
	  
  return FALSE;
}

static gboolean
init_list_of_services (void)
{
  struct dirent *dir;
  DIR *service_dir;
  gint i;
  GList *list;

  service_dir = opendir ("/etc/rc.d/rc5.d/");
  if (service_dir == NULL)
    return -1;
  while ((dir = readdir (service_dir)) != NULL)
    {
      if (check_for_invalid_service (dir->d_name))
	continue;

      services = g_list_prepend (services, g_strdup (dir->d_name));
    }
  closedir (service_dir);

  services = g_list_sort (services, (GCompareFunc)strcmp);

  /* now we stip off the leading 3 characters of each name */
  for (list = services; list; list = list->next)
    {
      gchar *service_name;

      service_name = g_strdup ((char *)list->data + 3);
      g_free (list->data);
      list->data = service_name;
    }

  for (i = G_N_ELEMENTS (system_service_names) - 1; i >= 0; i--)
    {
      services = g_list_prepend (services, g_strdup (system_service_names[i]));
    }

  n_intervals = g_list_length (services);

  return 0;
}


/* Display Window
 */

static void
create_main_window (int close_parent_fd)
{
  splash = splash_new ();

  close (close_parent_fd);
}

int
main (int argc, char *argv[])
{
  int pipe_fds[2] = { -1, -1 };
  int pid;
  int rc;
  poptContext context;

  const struct poptOption options[] = {
    POPT_AUTOHELP
    { "force", 'f', POPT_ARG_NONE, &force_arg, 1,
      N_("Force the graphical boot to start"), NULL },
    { "debug", 'd', POPT_ARG_NONE, &debug, 0,
      N_("leave debugging information up"), NULL },
    { "use-existing-server", 'u', POPT_ARG_STRING, &use_existing_server, 0,
      N_("use an existing server instead of creating one.  used primarily for debugging"), NULL },
    { "debug-interface", 'i', POPT_ARG_NONE, &debug_interface, 0,
      N_("Debug the graphical interface only"), NULL},
    { 0, 0, 0, 0, 0, 0, 0 }
  };

  setlocale (LC_ALL, "");

  bindtextdomain (GETTEXT_PACKAGE, "/usr/share/locale");
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
    
  context =  poptGetContext ("rhgb", argc, (const char **)argv,
			     options, 0);

  while ((rc = poptGetNextOpt (context)) > 0);
  if (rc < -1) {
    fprintf(stderr, "%s: %s\n",
	    poptBadOption(context, POPT_BADOPTION_NOALIAS),
	    poptStrerror(rc));
    return -1;
  }

  if (debug_interface)
    {
      DEBUG_MESSAGE("debug interface run\n");
      gtk_init_check (&argc, &argv);
      create_main_window (pipe_fds[1]);
      gtk_main ();
      exit (0);
    }

  DEBUG_MESSAGE("checking for X\n");
  if (check_for_x () != 0)
    setup_failed ();
  
  DEBUG_MESSAGE("checking for run\n");
  if (check_for_run () != 0 && !force_arg)
    setup_failed ();

  DEBUG_MESSAGE("create pipe\n");
  if (pipe (pipe_fds) != 0)
    setup_failed ();

  DEBUG_MESSAGE("create ramfs\n");
  if (setup_ramfs () != 0)
    setup_failed ();
  
  DEBUG_MESSAGE("forking\n");
  pid = fork ();
  if (pid < 0)
    setup_failed ();
  else if (pid > 0)
    {
      char *buf[5];
      close (pipe_fds[1]);
      read (pipe_fds[0], buf, 4);
      exit (0);
    }

  /* Set ourselves up as a daemon */
  close (pipe_fds[0]);

  if (! debug)
    {
      dup2 (open("/dev/null", O_RDWR), STDIN_FILENO);
      dup2 (open("/dev/null", O_RDWR), STDOUT_FILENO);
      dup2 (open("/dev/null", O_RDWR), STDERR_FILENO);
    }
  
  setsid ();
  setpgid (0, 0);

  /* try to cleanup in case of disaster */
  signal(SIGFPE, got_signal);
  signal(SIGPIPE, got_signal);
  signal(SIGILL, got_signal);
  signal(SIGSEGV, got_signal);

  DEBUG_MESSAGE("Initializing display\n");
  gtk_init_check (&argc, &argv);
  display_init ();

  DEBUG_MESSAGE("Initializing socket\n");
  if (socket_init () != 0)
    setup_failed ();

  if (init_list_of_services () != 0)
    setup_failed ();

  DEBUG_MESSAGE("Checking gtk\n");
  if (gtk_init_check (&argc, &argv) == 0)
	  setup_failed ();
 
  DEBUG_MESSAGE("Checking gtk okay\n");

  if (check_for_xrdb())
      merge_resources();

  DEBUG_MESSAGE("Creating main window\n");
  create_main_window (pipe_fds[1]);

  DEBUG_MESSAGE("Initializing pipe\n");
  if (pipe_init() != 0) {
    system(FIX_INIT_COMMAND);
    setup_failed ();
  }

  init_timeout();

  DEBUG_MESSAGE("Running GTK main loop\n");

  g_atexit (rhgb_shutdown);
  gtk_main ();

  DEBUG_MESSAGE("Shutdown\n");

  return exit_code;



}
