/* Copyright (c) Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gtk/gtk.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <popt.h>
#include <unistd.h>
#include <fcntl.h>
#include <locale.h>
#include <libintl.h>
#include <stdlib.h>

/**
 * ATTRIBUTE_UNUSED:
 *
 * Macro used to signal to GCC unused function parameters
 */
#ifdef __GNUC__
#ifdef HAVE_ANSIDECL_H
#include <ansidecl.h>
#endif
#ifndef ATTRIBUTE_UNUSED
#define ATTRIBUTE_UNUSED __attribute__((unused))
#endif
#else
#define ATTRIBUTE_UNUSED
#endif

/* Keep in sync w/ main.c */
#define SOCKET_NAME "/etc/rhgb/temp/rhgb-socket"
#define N_(x) x

static gboolean quit = 0;
static gboolean ping = 0;
static gboolean sysinit = 0;
char *details = NULL;
char *update = NULL;
static int socket_fd = 0;

static gint
socket_init (void)
{
  int curstate = 1;
  struct sockaddr_un *saddr; 

  socket_fd = socket (PF_UNIX, SOCK_STREAM, 0);
  fcntl (socket_fd, O_NONBLOCK);
    
  if (socket_fd < 0)
    return -1;

  if (setsockopt (socket_fd, SOL_SOCKET, SO_REUSEADDR, &curstate, sizeof(curstate)) < 0)
    return -1;

  saddr = g_new0 (struct sockaddr_un, 1);
  saddr->sun_family = AF_UNIX;
  memcpy (saddr->sun_path, SOCKET_NAME, sizeof (SOCKET_NAME));

  if (connect (socket_fd, (struct sockaddr *) saddr, sizeof (struct sockaddr_un)))
    return -1;

  return 0;
}

static void
handle_update (int sock_fd, char *message)
{
  char *real_message;

  real_message = g_strdup_printf ("update=%s", message);
  write (sock_fd, real_message, strlen (real_message) + 1);
  g_free (real_message);
}

static gboolean
quit_watcher (GIOChannel   *source,
	      GIOCondition  condition ATTRIBUTE_UNUSED,
	      gpointer      data)
{
  gchar buffer[20];
  gsize bytes_read = 0;
  GIOStatus status;
  GMainLoop *loop;

  loop = (GMainLoop *) data;

  status = g_io_channel_read_chars (source,
				    buffer,
				    20,
				    &bytes_read,
				    NULL);
  switch (status)
    {
    case G_IO_STATUS_NORMAL:
      if (bytes_read >= 20)
	goto done;
      if (strcmp (buffer, "done"))
	g_main_loop_quit (loop);
      return TRUE;
    case G_IO_STATUS_AGAIN:
      return TRUE;
    case G_IO_STATUS_EOF:
    case G_IO_STATUS_ERROR:
      break;
    default:
      g_assert_not_reached ();
    }

 done:
  g_main_loop_quit (loop);
  close (g_io_channel_unix_get_fd (source));
  return FALSE;

}

static gboolean
pong_watcher (GIOChannel   *source,
	      GIOCondition  condition ATTRIBUTE_UNUSED,
	      gpointer      data)
{
  gchar buffer[20];
  gsize bytes_read = 0;
  GIOStatus status;
  GMainLoop *loop;

  loop = (GMainLoop *) data;

  status = g_io_channel_read_chars (source,
				    buffer,
				    20,
				    &bytes_read,
				    NULL);
  switch (status)
    {
    case G_IO_STATUS_NORMAL:
      if (bytes_read >= 20)
	goto done;
      if (! strcmp (buffer, "pong"))
	exit (0);
      return TRUE;
    case G_IO_STATUS_AGAIN:
      return TRUE;
    case G_IO_STATUS_EOF:
    case G_IO_STATUS_ERROR:
      break;
    default:
      g_assert_not_reached ();
    }

 done:
  g_main_loop_quit (loop);
  close (g_io_channel_unix_get_fd (source));
  return FALSE;

}

static void
handle_quit (int sock_fd)
{
  GIOChannel *new_channel;
  GMainLoop *loop;

  if (write (sock_fd, "quit", strlen ("quit") + 1) == -1)
    exit (1);

  new_channel = g_io_channel_unix_new (sock_fd);
  loop = g_main_loop_new (NULL, FALSE);

  g_io_channel_set_encoding (new_channel, NULL, NULL);
  g_io_add_watch (new_channel, G_IO_IN, quit_watcher, loop);
  g_io_channel_unref (new_channel);

  g_main_loop_run (loop);
} 

static void
handle_details (int sock_fd, gchar *detail)
{
  if (! strcmp (detail, "yes"))
    {
      if (write (sock_fd, "details", strlen ("details") + 1) == -1)
	exit (1);
    }
  else 
    {
      if (write (sock_fd, "nodetails", strlen ("nodetails") + 1) == -1)
	exit (1);
    }

  exit (0);
}

static void
handle_sysinit (int sock_fd)
{
  if (write (sock_fd, "sysinit", strlen ("sysinit") + 1) == -1)
    exit (1);

  exit (0);
}



static void
handle_ping (int sock_fd)
{
  GIOChannel *new_channel;
  GMainLoop *loop;

  if (write (sock_fd, "ping", strlen ("ping") + 1) == -1)
    exit (1);

  new_channel = g_io_channel_unix_new (sock_fd);
  loop = g_main_loop_new (NULL, FALSE);

  g_io_channel_set_encoding (new_channel, NULL, NULL);
  g_io_add_watch (new_channel, G_IO_IN, pong_watcher, loop);
  g_io_channel_unref (new_channel);

  g_main_loop_run (loop);

  exit (1);
} 

int
main (int argc, const char *argv[])
{
    int rc;
    poptContext context;
    const struct poptOption options[] = {
	POPT_AUTOHELP
	{ "update", 'u', POPT_ARG_STRING, &update, 1,
	  N_("Update a service's status"), NULL },
	{ "details", 'd', POPT_ARG_STRING, &details, 1,
	  N_("Show the details page (yes/no)."), NULL },
	{ "ping", 'p', POPT_ARG_NONE, &ping, 0,
	  N_("See if the server is alive"), NULL },
	{ "quit", 'q', POPT_ARG_NONE, &quit, 0,
	  N_("Tells the server to quit"), NULL },
	{ "sysinit", 's', POPT_ARG_NONE, &sysinit, 0,
	  N_("Inform the server that we've finished rc.sysinit"), NULL },
	{ 0, 0, 0, 0, 0, 0, 0 }
    };

    
    setlocale(LC_ALL, "");
    bindtextdomain("rhgb-client", "/usr/share/locale");
    textdomain("rhgb-client");
    
    context =  poptGetContext ("rhgb-client",
			       argc, argv,
			       options, 0);

    while ((rc = poptGetNextOpt (context)) > 0);
    if (rc < -1) {
	fprintf(stderr, "%s: %s\n",
		poptBadOption(context, POPT_BADOPTION_NOALIAS),
		poptStrerror(rc));
	return -1;
    }
    
    if (!quit && !update && !ping && !details && ! sysinit)
      {
	fprintf(stderr, "Error: at least one argument must be specified\n");
	poptPrintHelp(context, stderr, 0);
	return -1;
      }

    if (details)
      {
	if (strcmp (details, "yes") && strcmp (details, "no"))
	  {
	    fprintf(stderr, "Error: details must be 'yes' or 'no'\n\n");
	    poptPrintHelp(context, stderr, 0);
	    return -1;
	  };
      }

    if (socket_init () != 0)
	return -1;

    if (quit)
      handle_quit (socket_fd);
    else if (details)
      handle_details (socket_fd, details);
    else if (ping)
      handle_ping (socket_fd);
    else if (update)
      handle_update (socket_fd, update);
    else if (sysinit)
      handle_sysinit (socket_fd);
    return 0;
}
